/*
 *        spThread.c
 */

#include <stdio.h>
#include <stdlib.h>

#include <sp/spBase.h>
#include <sp/spOption.h>
#include <sp/spThread.h>
#include <sp/spFile.h>

#if defined(MACOS) && (!defined(MACOSX64) && !defined(IPHONE))

#if defined(MACOSX)
#include <pthread.h>
#include <sys/time.h>
#if defined(IPHONE)
#include <CoreFoundation/CoreFoundation.h>
#else
#include <CoreServices/CoreServices.h>
#endif
#else
#include <Gestalt.h>
#include <Threads.h>
#endif

static int sp_num_coop_thread = 0;

typedef struct _spThread *spThread;
struct _spThread {
    ThreadID id;
    long return_status;
    volatile spBool end_flag;
    
    spThreadFunc func;
    void *data;
};
typedef struct _spMutex *spMutex;
struct _spMutex {
    ThreadID owner_thread;
    void *handle;
};

#elif defined(HAVE_PTHREAD)
#include <pthread.h>
#include <sys/time.h>
#include <errno.h>

#elif defined(_WIN32)
#include <windows.h>
#if defined(HAVE_BEGINTHREADEX)
#include <process.h>
#endif

typedef struct _spThread *spThread;
struct _spThread {
    void *handle;
    DWORD id;
};

static long waitThreadWin(void *handle)
{
    long status = 0;
    
    if (handle == NULL) return SP_THREAD_EXIT_FAILURE;
    
    {
        DWORD flag;
        DWORD nobject;
        MSG msg; 

        nobject = 1;
        while (1) {
            if ((flag = MsgWaitForMultipleObjects(nobject, &handle, 
                          FALSE, INFINITE, QS_ALLINPUT)) == WAIT_TIMEOUT) {
                return SP_THREAD_EXIT_FAILURE;
            }
            spDebug(80, "waitThreadWin", "flag = %ld (WAIT_OBJECT_0 = %ld, WAIT_ABANDONED_0 = %ld)\n",
                    flag, WAIT_OBJECT_0, WAIT_ABANDONED_0);

            if (flag == (WAIT_OBJECT_0 + nobject)) {
                /* received message */
                while (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE)) { 
                    spDebug(100, "waitThreadWin", "in message loop: message = %d\n", msg.message);
                    TranslateMessage(&msg);
                    DispatchMessage(&msg); 
                }
                continue;
            } else if (flag >= WAIT_OBJECT_0 && flag < WAIT_OBJECT_0 + nobject) {
                /* object is signaled */
                if (!GetExitCodeThread((HANDLE)handle, &flag)) {
                    return SP_THREAD_EXIT_FAILURE;
                }
                status = (long)flag;
                break;
            } else if (flag >= WAIT_ABANDONED_0 && flag < WAIT_ABANDONED_0 + nobject) {
                /* abandoned */
                status = -1;
                break;
            } else {
                /* other errors */
                return SP_THREAD_EXIT_FAILURE;
            }
        }
    }
    
    spDebug(80, "waitThreadWin", "done: status = %ld\n", status);

    return status;
}
#endif

#if defined(HAVE_PTHREAD) || defined(MACOSX)
static void *createThreadPT(long stacksize, int priority, spThreadFunc func, void *data)
{
    pthread_attr_t attr;
    pthread_t *thread;
    
    pthread_attr_init(&attr);
#ifndef linux
    {
        size_t size;
        pthread_attr_getstacksize (&attr, &size);
        if (size < stacksize) {
            pthread_attr_setstacksize (&attr, size);
        }
    }
#endif

    thread = malloc(sizeof(pthread_t));
    
    if (pthread_create(thread, &attr, func, data) != 0) {
        free(thread);
        thread = NULL;
    }
    
    pthread_attr_destroy(&attr);

    spDebug(50, "createThreadPT", "thread created: thread = %ld\n", (long)thread);
    
    return (void *)thread;
}

static void destroyThreadPT(void *handle)
{
    if (handle != NULL) {
        pthread_t *thread = (pthread_t *)handle;

        if (*thread != 0) {
            spDebug(50, "destroyThreadPT", "call pthread_detach for thread %ld\n", (long)*thread);
            pthread_detach(*thread);
        }
        free(handle);
    }
    
    return;
}

static void exitThreadPT(long status)
{
    pthread_exit((void *)status);
    return;
}

static void yieldThreadPT(void)
{
#if !defined(ANDROID)
    pthread_testcancel();
#endif
    return;
}

static long waitThreadPT(void *handle)
{
    long status = 0;
    void *ptr;
    int ret;
    pthread_t *thread;
    
    thread = (pthread_t *)handle;
    spDebug(80, "waitThreadPT", "thread = %ld\n", (long)thread);
    
    if ((ret = pthread_join(*thread, &ptr)) != 0) {
        spDebug(20, "waitThreadPT", "pthread_join for thread %ld returns nonzero: %d\n", (long)thread, ret);
        *thread = 0;
        return 1;
    }
    if (ptr == (void *)-1) {
        status = SP_THREAD_EXIT_CANCELED;
    } else {
        status = (long)ptr;
    }
    *thread = 0;

    spDebug(100, "waitThreadPT", "done: status = %ld\n", status);
    
    return status;
}

static spThreadId getThreadIdPT(void *handle)
{
    pthread_t *thread;

    thread = (pthread_t *)handle;
    
    return (spThreadId)*thread;
}

static spThreadId getCurrentThreadIdPT(void)
{
    return (spThreadId)pthread_self();
}

static spBool eqThreadIdPT(spThreadId id1, spThreadId id2)
{
    return spIsTrue(pthread_equal((pthread_t)id1, (pthread_t)id2));
}

void *createMutexPT(const char *name)
{
    pthread_mutex_t *mutex;
        
    mutex = malloc(sizeof(pthread_mutex_t));

    pthread_mutex_init(mutex, NULL);

    return (void *)mutex;
}

void destroyMutexPT(void *handle)
{
    pthread_mutex_destroy((pthread_mutex_t *)handle);
    free(handle);
    
    return;
}

spBool lockMutexPT(void *handle)
{
    if (pthread_mutex_lock((pthread_mutex_t *)handle) != 0) {
        return SP_FALSE;
    } else {
        return SP_TRUE;
    }
}

spBool unlockMutexPT(void *handle)
{
    if (pthread_mutex_unlock((pthread_mutex_t *)handle) != 0) {
        return SP_FALSE;
    } else {
        return SP_TRUE;
    }
}
#endif

#if defined(MACOS) && (!defined(MACOSX64) && !defined(IPHONE))
static spThreadReturn threadFuncMac(void *data)
{
    spThread thread = (spThread)data;
    spThreadReturn retval = NULL;

    if (thread != NULL && thread->func != NULL) {
        sp_num_coop_thread++;
    
        retval = thread->func(thread->data);
        thread->end_flag = SP_TRUE;
    
        sp_num_coop_thread = MAX(sp_num_coop_thread - 1, 0);
    }
    
    return retval;
}

int spGetNumCoopThreadMac(void)
{
    return sp_num_coop_thread;
}

static void *createThreadMac(long stacksize, int priority, spThreadFunc func, void *data)
{
    spThread thread;
    spThreadFunc tfunc;
    void *tdata;

#if !defined(MACOSX64)
    {
        SInt32 response;
        if (Gestalt(gestaltThreadMgrAttr, &response) != noErr
            || !(response & (1 << gestaltThreadMgrPresent))) {
            spDebug(1, "spCreateThread", "thread is not supported.\n");
            return NULL;
        }
    }
#endif

    thread = malloc(sizeof(struct _spThread));
    thread->id = 0;
    thread->return_status = 0;
    thread->end_flag = SP_FALSE;
    if (1) {
        thread->func = func;
        thread->data = data;
        tfunc = threadFuncMac;
        tdata = thread;
    } else {
        tfunc = func;
        tdata = data;
    }
    
    if (NewThread(kCooperativeThread,
                  /*(ThreadEntryProcPtr)*/(ThreadEntryTPP)tfunc, tdata,
                  stacksize, 0,
                  (void **)&thread->return_status, &thread->id) != noErr) {
        spDebug(1, "spCreateThread", "can't create thread.\n");
        free(thread);
        return NULL;
    }

    spDebug(50, "createThreadMac", "thread created: id = %ld\n", thread->id);
    
    return (void *)thread;
}

static void destroyThreadMac(void *handle)
{
    free(handle);
}

static void exitThreadMac(long status)
{
    DisposeThread(kCurrentThreadID, (void *)status, true);
    sp_num_coop_thread = MAX(sp_num_coop_thread - 1, 0);
    
    return;
}

static void yieldThreadMac(void)
{
    YieldToAnyThread();
    
    return;
}

static spThreadId getThreadIdMac(void *handle)
{
    spThread thread = (spThread)handle;

    return (spThreadId)thread->id;
}

static spThreadId getCurrentThreadIdMac(void)
{
    ThreadID id;
    
    if (MacGetCurrentThread(&id) != noErr) {
        return (spThreadId)kNoThreadID;
    }
    
    return (spThreadId)id;
}

static spMutex createMutexMac(const char *name)
{
    spMutex mutex;
    
    mutex = malloc(sizeof(struct _spMutex));
    mutex->owner_thread = kNoThreadID;
    mutex->handle = NULL;

#if defined(MACOSX)
    if (spIsLowLevelThreadUsed() == SP_TRUE) {
        mutex->handle = createMutexPT(name);
    }
#endif
    return mutex;
}

static void destroyMutexMac(spMutex mutex)
{
#if defined(MACOSX)
    if (mutex->handle != NULL) {
        destroyMutexPT(mutex->handle);
    }
#endif
    free(mutex);
    
    return;
}

static spBool lockMutexMac(spMutex mutex, spBool wait_only)
{
#if 0 && defined(MACOS9)
    ThreadBeginCritical();
#else
#if defined(MACOSX)
    if (spIsLowLevelThreadUsed() == SP_TRUE) {
        if (mutex->handle == NULL) {
            return SP_FALSE;
        }
        lockMutexPT(mutex->handle);
        if (wait_only == SP_TRUE) {
            unlockMutexPT(mutex->handle);
        }
        return SP_TRUE;
    }
#endif
    {
        ThreadID current_thread = kNoThreadID;
        
        MacGetCurrentThread(&current_thread);
        
        while (mutex->owner_thread != current_thread
               && mutex->owner_thread != kNoThreadID
               && YieldToThread(mutex->owner_thread) == noErr) {
            spDebug(100, "lockMutexMac", "waiting...\n");
        
        }
        ThreadBeginCritical();
        mutex->owner_thread = kNoThreadID;
        if (wait_only == SP_FALSE) {
            /*MacGetCurrentThread(&mutex->owner_thread);*/
            mutex->owner_thread = current_thread;
        }
        ThreadEndCritical();
    
        spDebug(100, "lockMutexMac", "done\n");
    
        if (wait_only == SP_TRUE) {
            YieldToAnyThread();
        }
    }
#endif

    return SP_TRUE;
}

static spBool unlockMutexMac(spMutex mutex)
{
#if 0 && defined(MACOS9)
    ThreadEndCritical();
    YieldToAnyThread();
#else
#if defined(MACOSX)
    if (spIsLowLevelThreadUsed() == SP_TRUE) {
        if (mutex->handle == NULL) {
            return SP_FALSE;
        }
        unlockMutexPT(mutex->handle);
        return SP_TRUE;
    }
#endif
    
    ThreadBeginCritical();
    mutex->owner_thread = kNoThreadID;
    ThreadEndCritical();
    /*YieldToAnyThread();*/
#endif
    
    return SP_TRUE;
}

static long waitThreadMac(void *handle)
{
    spThread thread = (spThread)handle;
    ThreadState state;
    
    while (thread->end_flag == SP_FALSE) {
        spDebug(100, "waitThreadMac", "waiting thread %ld ...\n", thread->id);
        if (YieldToThread(thread->id) != noErr) {
            break;
        }
        state = kStoppedThreadState;
        if (GetThreadState(thread->id, &state) != noErr) {
            break;
        } else {
            spDebug(100, "waitThreadMac", "thread state = %d\n", state);
            if (state == kStoppedThreadState) {
                break;
            }
        }
    }

    spDebug(100, "waitThreadMac", "done: status = %ld\n", thread->return_status);
    
    return thread->return_status;
}
#endif

void *spCreateThread(long stacksize, int priority, spThreadFunc func, void *data)
{
#if defined(MACOS) && (!defined(MACOSX64) && !defined(IPHONE))
#if defined(MACOSX)
    if (spIsLowLevelThreadUsed() == SP_TRUE) {
        spDebug(50, "spCreateThread", "create thread using pthread\n");
        return createThreadPT(stacksize, priority, func, data);
    }
#endif
    return createThreadMac(stacksize, priority, func, data);
#elif defined(HAVE_PTHREAD)
    return createThreadPT(stacksize, priority, func, data);
#elif defined(_WIN32)
    spThread hthread;

    hthread = xalloc(1, struct _spThread);

#if defined(HAVE_BEGINTHREADEX)
    hthread->handle = (void *)_beginthreadex(NULL, stacksize, (LPTHREAD_START_ROUTINE)func, data,
                                             0, (unsigned *)&hthread->id);
#else
    hthread->handle = (void *)CreateThread(NULL, stacksize, (LPTHREAD_START_ROUTINE)func, data,
                                           0, &hthread->id);
#endif
    SetThreadPriority((HANDLE)hthread->handle, priority);
    
    return hthread;
#else
    return NULL;
#endif
}

void spDestroyThread(void *handle)
{
    if (handle == NULL) return;
    
#if defined(MACOS) && (!defined(MACOSX64) && !defined(IPHONE))
#if defined(MACOSX)
    if (spIsLowLevelThreadUsed() == SP_TRUE) {
        destroyThreadPT(handle);
        return;
    }
#endif
    destroyThreadMac(handle);
#elif defined(HAVE_PTHREAD)
    destroyThreadPT(handle);
#elif defined(_WIN32)
    {
        spThread hthread = (spThread)handle;
        
        CloseHandle((HANDLE)hthread->handle);
        xfree(hthread);
    }
#endif
    
    return;
}

void spExitThread(long status)
{
#if defined(MACOS) && (!defined(MACOSX64) && !defined(IPHONE))
#if defined(MACOSX)
    if (spIsLowLevelThreadUsed() == SP_TRUE) {
        exitThreadPT(status);
        return;
    }
#endif
    exitThreadMac(status);
#elif defined(HAVE_PTHREAD)
    exitThreadPT(status);
#elif defined(_WIN32)
#if defined(HAVE_BEGINTHREADEX)
    _endthreadex((unsigned)status);
#else
    ExitThread((DWORD)status);
#endif
#endif
    
    return;
}

void spYieldThread(void)
{
#if defined(MACOS) && (!defined(MACOSX64) && !defined(IPHONE))
#if defined(MACOSX)
    if (spIsLowLevelThreadUsed() == SP_TRUE) {
        yieldThreadPT();
        return;
    }
#endif
    yieldThreadMac();
#elif defined(HAVE_PTHREAD)
    yieldThreadPT();
#endif
    return;
}

long spWaitThread(void *handle)
{
    long status = 0;
    
    if (handle == NULL) return SP_THREAD_EXIT_FAILURE;
    
#if defined(MACOS) && (!defined(MACOSX64) && !defined(IPHONE))
#if defined(MACOSX)
    if (spIsLowLevelThreadUsed() == SP_TRUE) {
        return waitThreadPT(handle);
    }
#endif
    status = waitThreadMac(handle);
#elif defined(HAVE_PTHREAD)
    status = waitThreadPT(handle);
#elif defined(_WIN32)
    {
        spThread hthread = (spThread)handle;
    
        status = waitThreadWin(hthread->handle);
    }
#endif

    return status;
}

spThreadId spGetThreadId(void *handle)
{
    spThreadId id;
    
#if defined(MACOS) && (!defined(MACOSX64) && !defined(IPHONE))
#if defined(MACOSX)
    if (spIsLowLevelThreadUsed() == SP_TRUE) {
        return getThreadIdPT(handle);
    }
#endif
    id = getThreadIdMac(handle);
#elif defined(HAVE_PTHREAD)
    id = getThreadIdPT(handle);
#elif defined(_WIN32)
    {
        spThread hthread = (spThread)handle;
        id = hthread->id;
    }
#endif

    return id;
}

spThreadId spGetCurrentThreadId(void)
{
    spThreadId id;
    
#if defined(MACOS) && (!defined(MACOSX64) && !defined(IPHONE))
#if defined(MACOSX)
    if (spIsLowLevelThreadUsed() == SP_TRUE) {
        return getCurrentThreadIdPT();
    }
#endif
    id = getCurrentThreadIdMac();
#elif defined(HAVE_PTHREAD)
    id = getCurrentThreadIdPT();
#elif defined(_WIN32)
    id = GetCurrentThreadId();
#endif

    return id;
}

spBool spEqThreadId(spThreadId id1, spThreadId id2)
{
    spBool ret;
    
#if defined(HAVE_PTHREAD)
#if defined(MACOSX) && (!defined(MACOSX64) && !defined(IPHONE))
    if (spIsLowLevelThreadUsed() == SP_FALSE) {
        return spIsTrue(id1 == id2);
    }
#endif
    ret = eqThreadIdPT(id1, id2);
#else
    ret = spIsTrue(id1 == id2);
#endif
    
    return ret;
}

void *spCreateMutex(const char *name)
{
    void *handle = NULL;
    
#if defined(MACOS) && (!defined(MACOSX64) && !defined(IPHONE))
    handle = (void *)createMutexMac(name);
#elif defined(HAVE_PTHREAD)
    handle = createMutexPT(name);
#elif defined(_WIN32)
    handle = (void *)CreateMutex(NULL, FALSE, name);
#endif

    return handle;
}

void spDestroyMutex(void *handle)
{
    if (handle == NULL) return;
    
#if defined(MACOS) && (!defined(MACOSX64) && !defined(IPHONE))
    destroyMutexMac((spMutex)handle);
#elif defined(HAVE_PTHREAD)
    destroyMutexPT(handle);
#elif defined(_WIN32)
    CloseHandle((HANDLE)handle);
#endif

    return;
}

spBool spLockMutex(void *handle)
{
    if (handle == NULL) return SP_FALSE;
    
#if defined(MACOS) && (!defined(MACOSX64) && !defined(IPHONE))
    if (lockMutexMac((spMutex)handle, SP_FALSE) == SP_FALSE) {
        return SP_FALSE;
    }
#elif defined(HAVE_PTHREAD)
    if (lockMutexPT(handle) == SP_FALSE) {
        return SP_FALSE;
    }
#elif defined(_WIN32)
    if (WaitForSingleObject((HANDLE)handle, INFINITE) == WAIT_TIMEOUT) {
        return SP_FALSE;
    }
#endif
    return SP_TRUE;
}

spBool spUnlockMutex(void *handle)
{
    if (handle == NULL) return SP_FALSE;
    
#if defined(MACOS) && (!defined(MACOSX64) && !defined(IPHONE))
    if (unlockMutexMac((spMutex)handle) == SP_FALSE) {
        return SP_FALSE;
    }
#elif defined(HAVE_PTHREAD)
    if (unlockMutexPT(handle) == SP_FALSE) {
        return SP_FALSE;
    }
#elif defined(_WIN32)
    ReleaseMutex((HANDLE)handle);
#endif
    return SP_TRUE;
}

spBool spWaitMutex(void *handle)
{
#if defined(MACOS) && (!defined(MACOSX64) && !defined(IPHONE))
    if (handle == NULL) return SP_FALSE;
    
    return lockMutexMac((spMutex)handle, SP_TRUE);
#else
    if (spLockMutex(handle) == SP_TRUE) {
        spUnlockMutex(handle);
        return SP_TRUE;
    }
    
    return SP_FALSE;
#endif
}

#if !defined(_WIN32)
typedef struct _spEvent *spEvent;
struct _spEvent {
    volatile spBool signal_flag;
    spBool manual_reset;

    volatile int num_wait;
    
#if defined(HAVE_PTHREAD) || defined(MACOSX)
    pthread_cond_t cond;
    pthread_mutex_t mutex;
#endif
    volatile spBool pulsed;
    
#if defined(MACOS) && (!defined(MACOSX64) && !defined(IPHONE))
    ThreadID wait_thread;
#endif
};

static spEvent createEvent(spBool manual_reset, spBool initial_signal)
{
    spBool failed = SP_FALSE;
    spEvent event = NULL;

#if defined(HAVE_PTHREAD) || defined(MACOSX) || defined(MACOS)
    event = xalloc(1, struct _spEvent);
    memset(event, 0, sizeof(struct _spEvent));
    event->signal_flag = initial_signal;
    event->manual_reset = manual_reset;
    event->num_wait = 0;
#else
    {
        return NULL;
    }
#endif
        
#if defined(HAVE_PTHREAD) || defined(MACOSX)
    if (pthread_cond_init(&event->cond, NULL) != 0) {
        failed = SP_TRUE;
    } else {
        if (pthread_mutex_init(&event->mutex, NULL) != 0) {
            failed = SP_TRUE;
        }
    }
#endif
#if defined(MACOS) && (!defined(MACOSX64) && !defined(IPHONE))
    event->wait_thread = kNoThreadID;
#endif

    if (failed) {
        xfree(event);
        return NULL;
    }
    
    event->pulsed = SP_FALSE;

    return event;
}

static spBool destroyEvent(spEvent event)
{
    spBool flag = SP_FALSE;
    
    flag = SP_TRUE;

#if defined(HAVE_PTHREAD) || defined(MACOSX)
    if (pthread_mutex_destroy(&event->mutex) != 0) {
        flag = SP_FALSE;
    }
    if (pthread_cond_destroy(&event->cond) != 0) {
        flag = SP_FALSE;
    }
#endif
    
    xfree(event);

    return flag;
}

#if defined(HAVE_PTHREAD) || defined(MACOSX)
static int waitEventTimeoutPT(spEvent event, long timeout_ms)
{
    int ret = SP_EVENT_WAIT_FAILED;
    
    spDebug(20, "waitEventTimeoutPT", "wait...\n");
    
    if (pthread_mutex_lock(&event->mutex) == 0) {
        ++event->num_wait;
        spDebug(50, "waitEventTimeoutPT", "signal_flag = %d, num_wait = %d\n",
                event->signal_flag, event->num_wait);
        
        if (event->pulsed == SP_TRUE) {
            event->signal_flag = SP_FALSE;
            event->pulsed = SP_FALSE;
        }

        if (timeout_ms == SP_EVENT_TIMEOUT_INFINITE) {
            while (event->signal_flag == SP_FALSE) {
                spDebug(100, "waitEventTimeoutPT", "before pthread_cond_wait\n");
                pthread_cond_wait(&event->cond, &event->mutex);
                spDebug(100, "waitEventTimeoutPT", "after pthread_cond_wait\n");
            }
            ret = SP_EVENT_WAIT_SIGNALED;
        } else {
            int retcode;
            long timeout_s;
            long rem_ms;
            struct timeval now;
            struct timespec timeout;

            gettimeofday(&now, NULL);
            if (timeout_ms == SP_EVENT_TIMEOUT_INFINITE) {
                timeout_s = 0;
                rem_ms = 10;
            } else {
                timeout_s = (timeout_ms / 1000);
                rem_ms = timeout_ms - (timeout_s * 1000);
            }
            timeout.tv_sec = now.tv_sec + (time_t)timeout_s;
            timeout.tv_nsec = (now.tv_usec * 1000) + (rem_ms * 1000000);

            retcode = 0;
            while (event->signal_flag == SP_FALSE) {
                spDebug(100, "waitEventTimeoutPT", "before pthread_cond_wait\n");
                retcode = pthread_cond_timedwait(&event->cond, &event->mutex, &timeout);
                spDebug(100, "waitEventTimeoutPT", "after pthread_cond_wait: retcode = %d\n", retcode);
                if (timeout_ms != SP_EVENT_TIMEOUT_INFINITE && retcode == ETIMEDOUT) {
                    break;
                }
            }
            if (retcode == ETIMEDOUT) {
                ret = SP_EVENT_WAIT_TIMEOUT;
            } else {
                ret = SP_EVENT_WAIT_SIGNALED;
            }
        }
            
        if (event->manual_reset == SP_FALSE) {
            event->signal_flag = SP_FALSE;
        }
        if (event->pulsed == SP_TRUE) {
            event->signal_flag = SP_FALSE;
            event->pulsed = SP_FALSE;
        }
            
        --event->num_wait;
        event->num_wait = MAX(event->num_wait, 0);
        pthread_mutex_unlock(&event->mutex);
    }

    spDebug(20, "waitEventTimeoutPT", "done: ret = %d\n", ret);
    
    return ret;
}

static spBool setOrResetEventPT(spEvent event, spBool signal_flag)
{
    spBool flag = SP_FALSE;
    
    spDebug(20, "setOrResetEventPT", "in: signal_flag = %d\n", signal_flag);
    
    if (pthread_mutex_lock(&event->mutex) == 0) {
        event->pulsed = SP_FALSE;
        spDebug(20, "setOrResetEventPT", "original event->signal_flag = %d\n", event->signal_flag);
        if (signal_flag == SP_TRUE) {
            if (event->signal_flag == SP_FALSE) {
                event->signal_flag = signal_flag;
                spDebug(20, "setOrResetEventPT", "call pthread_cond_signal\n");
                pthread_cond_signal(&event->cond);
            }
        } else {
            event->signal_flag = signal_flag;
        }
        spDebug(20, "setOrResetEventPT", "updated event->signal_flag = %d\n", event->signal_flag);
        pthread_mutex_unlock(&event->mutex);
        flag = SP_TRUE;
    }

    spDebug(20, "setOrResetEventPT", "done: flag = %d\n", flag);
    
    return flag;
}

static spBool pulseEventPT(spEvent event)
{
    spBool flag = SP_FALSE;
    
    spDebug(20, "pulseEventPT", "in\n");
    
    if (pthread_mutex_lock(&event->mutex) == 0) {
        if (event->num_wait <= 0) {
            event->pulsed = SP_FALSE;
            event->signal_flag = SP_FALSE;
        } else {
            event->pulsed = SP_TRUE;
            event->signal_flag = SP_TRUE;
            spDebug(20, "pulseEventPT", "call pthread_cond_signal\n");
            pthread_cond_signal(&event->cond);
        }
        flag = SP_TRUE;
        pthread_mutex_unlock(&event->mutex);
    }

    spDebug(20, "pulseEventPT", "done: flag = %d\n", flag);
    
    return flag;
}
#endif

#if defined(MACOS) && (!defined(MACOSX64) && !defined(IPHONE))
static int waitEventTimeoutMac(spEvent event, long timeout_ms)
{
    int ret = SP_EVENT_WAIT_FAILED;
    int i;
    spBool wait_flag = SP_FALSE;
    ThreadID thread_id;
    
#if defined(MACOSX)
    if (spIsLowLevelThreadUsed() == SP_TRUE) {
        return waitEventTimeoutPT(event, timeout_ms);
    }
#endif
    
    ThreadBeginCritical();

    MacGetCurrentThread(&thread_id);
    
    if (event->wait_thread == kNoThreadID) {
        event->wait_thread = thread_id;
        wait_flag = SP_TRUE;
    }
    
    if (event->pulsed == SP_TRUE) {
        event->signal_flag = SP_FALSE;
        event->pulsed = SP_FALSE;
    }

    if (timeout_ms == SP_EVENT_TIMEOUT_INFINITE) {
        for (i = 0; event->signal_flag == SP_FALSE; i++) {
            ThreadEndCritical();
            spDebug(80, "waitEventTimeoutMac", "i = %d, before yield\n", i);
            YieldToAnyThread();
            spDebug(80, "waitEventTimeoutMac", "i = %d, after yield\n", i);
            ThreadBeginCritical();
            if (event->wait_thread == kNoThreadID) { /* another thread was released */
                event->wait_thread = thread_id;
                wait_flag = SP_TRUE;
            }
        }
        ret = SP_EVENT_WAIT_SIGNALED;
    } else {
        unsigned long init_tick, last_tick;
        unsigned long timeout_tick;
        
        timeout_tick = (60L * timeout_ms) / 1000L;
        init_tick = TickCount();
        
        ret = SP_EVENT_WAIT_SIGNALED;
        
        for (i = 0; event->signal_flag == SP_FALSE; i++) {
            ThreadEndCritical();
            YieldToAnyThread();
            ThreadBeginCritical();
            
            last_tick = TickCount();
            if (last_tick - init_tick >= timeout_tick) {
                ret = SP_EVENT_WAIT_TIMEOUT;
                break;
            } else if (event->wait_thread == kNoThreadID) { /* another thread was released */
                event->wait_thread = thread_id;
                wait_flag = SP_TRUE;
            }
        }
    }
        
    if (event->manual_reset == SP_FALSE) {
        event->signal_flag = SP_FALSE;
    }
    if (event->pulsed == SP_TRUE) {
        event->signal_flag = SP_FALSE;
        event->pulsed = SP_FALSE;
    }

    if (wait_flag == SP_TRUE) {
        event->wait_thread = kNoThreadID;
    }
            
    ThreadEndCritical();
    
    return ret;
}

static spBool setOrResetEventMac(spEvent event, spBool signal_flag)
{
    spBool flag = SP_FALSE;
    
#if defined(MACOSX)
    if (spIsLowLevelThreadUsed() == SP_TRUE) {
        return setOrResetEventPT(event, signal_flag);
    }
#endif
    
    ThreadBeginCritical();
    
    event->pulsed = SP_FALSE;
    event->signal_flag = signal_flag;
    flag = SP_TRUE;

    ThreadEndCritical();
    if (event->signal_flag == SP_TRUE) {
        if (event->wait_thread != kNoThreadID) {
            YieldToThread(event->wait_thread);
        } else {
            YieldToAnyThread();
        }
    }
    
    return flag;
}

static spBool pulseEventMac(spEvent event)
{
    spBool flag = SP_FALSE;
    
#if defined(MACOSX)
    if (spIsLowLevelThreadUsed() == SP_TRUE) {
        return pulseEventPT(event);
    }
#endif
    
    ThreadBeginCritical();
    if (event->num_wait <= 0) {
        event->pulsed = SP_FALSE;
        event->signal_flag = SP_FALSE;
    } else {
        event->pulsed = SP_TRUE;
        event->signal_flag = SP_TRUE;
    }
    flag = SP_TRUE;
    ThreadEndCritical();
    if (event->signal_flag == SP_TRUE) {
        if (event->manual_reset == SP_TRUE) {
            YieldToAnyThread();
        } else if (event->wait_thread != kNoThreadID) {
            YieldToThread(event->wait_thread);
        } else {
            /* do nothing */
        }
    }

    return flag;
}

#endif
#endif

void *spCreateEvent(spBool manual_reset, spBool initial_signal)
{
    void *handle = NULL;

#if defined(_WIN32)
    handle = CreateEvent(NULL, manual_reset, initial_signal, NULL);
#else
    handle = (void *)createEvent(manual_reset, initial_signal);
#endif

    return handle;
}

spBool spDestroyEvent(void *handle)
{
    spBool flag = SP_FALSE;
    
    if (handle == NULL) return SP_FALSE;

#if defined(_WIN32)
    if (CloseHandle(handle)) {
        flag = SP_TRUE;
    }
#else
    flag = destroyEvent((spEvent)handle);
#endif

    return flag;
}

int spWaitEventTimeout(void *handle, long timeout_ms)
{
    int ret = SP_EVENT_WAIT_FAILED;
    
    if (handle == NULL) return SP_EVENT_WAIT_FAILED;

#if defined(_WIN32)
    {
        DWORD dwTimeout;
        DWORD dwError;
        
        if (timeout_ms == SP_EVENT_TIMEOUT_INFINITE) {
            dwTimeout = INFINITE;
        } else {
            dwTimeout = (DWORD)MAX(timeout_ms, 0);
        }
        dwError = WaitForSingleObject((HANDLE)handle, dwTimeout);
        if (dwError == WAIT_OBJECT_0) {
            ret = SP_EVENT_WAIT_SIGNALED;
        } else if (dwError == WAIT_TIMEOUT) {
            ret = SP_EVENT_WAIT_TIMEOUT;
        } else if (dwError == WAIT_ABANDONED) {
            ret = SP_EVENT_WAIT_ABANDONED;
        }
    }
#elif defined(MACOS) && (!defined(MACOSX64) && !defined(IPHONE))
    ret = waitEventTimeoutMac((spEvent)handle, timeout_ms);
#elif defined(HAVE_PTHREAD)
    ret = waitEventTimeoutPT((spEvent)handle, timeout_ms);
#endif

    return ret;
}

spBool spWaitEvent(void *handle)
{
    int ret;
    
    ret = spWaitEventTimeout(handle, SP_EVENT_TIMEOUT_INFINITE);

    if (ret == SP_EVENT_WAIT_SIGNALED) {
        return SP_TRUE;
    } else {
        return SP_FALSE;
    }
}

static spBool setOrResetEvent(void *handle, spBool signal_flag)
{
    spBool flag = SP_FALSE;
    
    if (handle == NULL) return SP_FALSE;
        
#if defined(_WIN32)
    if (signal_flag == SP_TRUE) {
        if (SetEvent((HANDLE)handle)) {
            flag = SP_TRUE;
        }
    } else {
        if (ResetEvent((HANDLE)handle)) {
            flag = SP_TRUE;
        }
    }
#elif defined(MACOS) && (!defined(MACOSX64) && !defined(IPHONE))
    flag = setOrResetEventMac((spEvent)handle, signal_flag);
#elif defined(HAVE_PTHREAD)
    flag = setOrResetEventPT((spEvent)handle, signal_flag);
#endif

    return flag;
}

spBool spSetEvent(void *handle)
{
    return setOrResetEvent(handle, SP_TRUE);
}

spBool spResetEvent(void *handle)
{
    return setOrResetEvent(handle, SP_FALSE);
}

spBool spPulseEvent(void *handle)
{
    spBool flag = SP_FALSE;

    if (handle == NULL) return SP_FALSE;
    
#if defined(_WIN32)
    if (PulseEvent((HANDLE)handle)) {
        flag = SP_TRUE;
    }
#elif defined(MACOS) && (!defined(MACOSX64) && !defined(IPHONE))
    flag = pulseEventMac((spEvent)handle);
#elif defined(HAVE_PTHREAD)
    flag = pulseEventPT((spEvent)handle);
#endif

    return flag;
}

#if 0
/* semaphore */
#if defined(_WIN32)
#elif defined(HAVE_PTHREAD)
#endif
#endif

#if defined(MACOS) && (!defined(MACOSX64) && !defined(IPHONE))
static spThreadLevel sp_thread_level = SP_THREAD_LEVEL_HIGH;
#else
static spThreadLevel sp_thread_level = SP_THREAD_LEVEL_LOW;
#endif

spBool spIsThreadLevelVariable(void)
{
#if defined(MACOSX)
    return SP_TRUE;
#else
    return SP_FALSE;
#endif
}

spBool spIsLowLevelThreadUsed(void)
{
    if (sp_thread_level == SP_THREAD_LEVEL_LOW) {
        return SP_TRUE;
    } else {
        return SP_FALSE;
    }
}

spBool spSetThreadLevel(spThreadLevel level)
{
#if defined(MACOSX)
    if (level == SP_THREAD_LEVEL_LOW || level == SP_THREAD_LEVEL_HIGH) {
        sp_thread_level = level;
        return SP_TRUE;
    }
#endif
    return SP_FALSE;
}

spThreadLevel spGetThreadLevel(void)
{
    return sp_thread_level;
}

#if !defined(_WIN32) && !defined(MACOS)
#include <errno.h>
#endif

typedef struct _spCommandThreadData
{
    char *command;
    spCommandThreadCallback func;
    void *data;
} *spCommandThreadData;

#if !defined(IPHONE)
static spThreadReturn commandThread(void *data)
{
    long err;
    spCommandThreadData thread_data = (spCommandThreadData)data;
#if defined(MACOS9)
    int errno;
#endif
    
    if (thread_data->func != NULL) thread_data->func(SP_COMMAND_THREAD_STARTED, 0, thread_data->data);
    
#if defined(_WIN32)
    {
        DWORD flag;
        STARTUPINFO si;
        PROCESS_INFORMATION pi;

        ZeroMemory(&si, sizeof( STARTUPINFO));
        si.cb = sizeof( STARTUPINFO );
        si.dwFlags = STARTF_USESHOWWINDOW;
        si.wShowWindow = SW_SHOWNORMAL;
        
        if (CreateProcess(NULL, thread_data->command,
                          NULL, NULL, FALSE, 0, NULL, NULL, &si, &pi)) {
            CloseHandle(pi.hThread);

            if ((flag = WaitForSingleObject(pi.hProcess, INFINITE)) != WAIT_OBJECT_0) {
                /* error */
                err = (long)flag;
                if (thread_data->func != NULL) thread_data->func(SP_COMMAND_THREAD_COMMAND_WAIT_FAILED, err, thread_data->data);
            } else {
                if (!GetExitCodeThread(pi.hProcess, &flag)) {
                    flag = -1;
                }
                err = (long)flag;
                
                if (thread_data->func != NULL) thread_data->func(SP_COMMAND_THREAD_COMMAND_FINISHED, err, thread_data->data);
            }
            
            
            CloseHandle(pi.hProcess);
        } else {
            err = (long)GetLastError();
            if (thread_data->func != NULL) thread_data->func(SP_COMMAND_THREAD_COMMAND_FAILED, err, thread_data->data);
        }
    }
#else /* defined(_WIN32) */
    errno = 0;
    err = system(thread_data->command);
    if (errno != 0 && err != 0) {
        if (thread_data->func != NULL) thread_data->func(SP_COMMAND_THREAD_COMMAND_FAILED, err, thread_data->data);
    } else {
        if (thread_data->func != NULL) thread_data->func(SP_COMMAND_THREAD_COMMAND_FINISHED, err, thread_data->data);
    }
#endif /* !defined(_WIN32) */

    if (thread_data->func != NULL) thread_data->func(SP_COMMAND_THREAD_FINISHED, 0, thread_data->data);

    xfree(thread_data->command);
    xfree(thread_data);
    
    return SP_THREAD_RETURN_SUCCESS;
}

static void commandThreadExitCallback(void *hthread)
{
    if (hthread != NULL) {
        spDebug(1, "commandThreadExitCallback", "command thread %ld is destroyed\n",
                (long)hthread);
        spDestroyThread(hthread);
    }

    return;
}
#endif /* !defined(IPHONE) */

spBool spCreateCommandThread(const char *command, spCommandThreadCallback func, void *data)
{
    void *hthread;
    spCommandThreadData thread_data;

    if (strnone(command)) return SP_FALSE;

#if defined(IPHONE)
    return SP_FALSE;
#else
    spDebug(10, "spCreateCommandThread", "command = %s\n", command);
    
    thread_data = xalloc(1, struct _spCommandThreadData);
#if defined(MACOSX)
    {
        char *args;
        char buf[SP_MAX_PATHNAME];

        thread_data->command = NULL;

        if ((args = sgetnextncol(buf, SP_MAX_PATHNAME, command)) != 0) {
            spDebug(10, "spCreateCommandThread", "buf = %s\n", buf);
            
            if ((spEqSuffix(buf, ".app") == SP_TRUE
                 || spReplaceSuffix(buf, ".app") == SP_TRUE)
                && spExists(buf) == SP_TRUE) {
                char *exename;
                
                exename = xspGetBaseName(buf);
                spRemoveSuffix(exename, NULL);
                spDebug(10, "spCreateCommandThread", "exename = %s\n", exename);

                thread_data->command = xalloc(strlen(buf) + strlen(exename)
                                              + strlen(args) + 20, char);
                sprintf(thread_data->command, "\"%s/Contents/MacOS/%s\" %s", buf, exename, args);
                spDebug(10, "spCreateCommandThread", "command = %s\n", thread_data->command);
                
                xfree(exename);
            }
        }

        if (thread_data->command == NULL) {
            thread_data->command = strclone(command);
        }
    }
#else /* defined(MACOSX) */
    thread_data->command = strclone(command);
#endif /* !defined(MACOSX) */
    thread_data->func = func;
    thread_data->data = data;

    if ((hthread = spCreateThread(0, 0, commandThread, thread_data)) == NULL) {
        return SP_FALSE;
    }
    spAddExitCallback(commandThreadExitCallback, hthread);

    spDebug(10, "spCreateCommandThread", "done\n");
    
    return SP_TRUE;
#endif /* !defined(IPHONE) */
}
